/* 
 *  $Id: ConfigurationManager.Java,v 1.8 1999/10/29 00:53:56 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_ANTS_UTILS_PKG.KeyArgs;
import JANOS_ANTS_UTILS_PKG.InvalidKeyArgsException;
import JANOS_ANTS_UTILS_PKG.LineListTokenizer;
import JANOS_ANTS_UTILS_PKG.LineListException;

import JANOS_CORE_PKG.Address;
import JANOS_CORE_PKG.AddressFormatException;
import JANOS_CORE_PKG.BufferHandle;
import JANOS_CORE_PKG.CommSpaceHandle;
import JANOS_CORE_PKG.Flow;
import JANOS_CORE_PKG.FlowRun;
import JANOS_CORE_PKG.ThreadPool;
import JANOS_CORE_PKG.NoSuchNodeAddress;

import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Vector;

IMPORT_DEBUG
IMPORT_TRACE

/**
 * A configuration manager that creates a node interface,
 * channel and app configuration.  Gets the configuration
 * from a control file.
 * <p>
 * Really, these functions could go on the Node class.
 * <p>
 * XXX this shouldn't be public, or arbitrary Capsules
 * can it up...
 *
 * %AUTHOR_PAT%
 * @author David Wetherall
 */
public class ConfigurationManager
	extends FlowRun
{
	private Flow flow;
	private ThreadPool threadPool;

	/**
	 * Not private so that the inner Runnable() created on the
	 * channel command can see it and use it for the thread it
	 * starts up on the main message loop.
	 */
	PACKAGE static int port = 4242;

	/**
	 * Parse a single command.
	 * 
	 * @param node the node to attach created entities to
	 * @param command array of command and its argument words
	 * @exception AddressFormatException if an address parameter is bad
	 * @exception ArrayIndexOutOfBoundsException if insufficent paramters are given for some commands.
	 * @exception AttachException if attaching Channel/Extension/App fails
	 * @exception ClassNotFoundException if any dynamically loaded class isn't found.
	 * @exception IllegalAccessException if newInstance() fails
	 * @exception InstantiationException if newInstance() fails
	 * @exception InvalidKeyArgsException if any optional args are improperly formatted
	 */
	private final void parseCommandLine(Vector appVector, String[] command)
		throws 
		  AddressFormatException,
		  ArrayIndexOutOfBoundsException,
		  AttachException,
		  ClassNotFoundException,
		  IllegalAccessException,
		  InstantiationException,
		  InvalidKeyArgsException
	{
		TRACE(Trace.ANTS_STARTUP, "Parsing command " +command[0]);

		/*
		 * Format of all command lines:
		 * 	<command> <address> [<args> ...]
		 * The first two elements are required.  The args are dependent
		 * upon the command.
		 */
		if (command[0].equals("node")) {
			// FORMAT: node <node addr> [<optional arg> ...]

			// XXX node.setName("Node " + command[1]);

			// Set address before calling setArgs so that routeTable will exist.
			PrimordialNode.setAddress(Address.fromString(command[1]));
			PrimordialNode.setArgs(new KeyArgs(command, 2));

		} else if (command[0].equals("channel")) {
			// FORMAT: channel <node addr> <physical addr> [<optional args> ... ]

			if (command.length < 3)
				throw new ArrayIndexOutOfBoundsException("insufficent "
									 + "arguments for "
									 + "channel command");

			String physAddr = command[2];
			String host;
			int sep = physAddr.indexOf(':');
			if (sep != 0)
			{
				try
				{
					host = physAddr.substring(0,sep);
					assert(host.equals("localhost")); // XXX
					port = Integer.parseInt(physAddr.substring(sep+1,physAddr.length()));
				// above might throw NumberFormatException
				}
				catch (Exception e)
				{
					panic("Problem parsing channel address: " +e.getMessage());
				}
			}

			// XXX exactly one channel per node (only one physical address per node!)
			PrimordialNode.setPhysicalAddress(physAddr);

			// XXX Fire up a new Thread to run the receive message loop... XXX

			Runnable r = new Runnable() {
				public void run() {
					try
					{
						JANOS_CORE_PKG.InChannel.msgLoop(port);
					}
					catch (IOException ioe)
					{
						FATALEXCEPTION(ioe, "MSGLOOP THREAD GOT ALL CONFUSED");
					}
				}
			};
			
			threadPool.threadStart(r);
			
		} else if (command[0].equals("application")) {
			// FORMAT: application <node addr> <app class> [<optional args> ...]
			
			if (command.length < 3)
				throw new ArrayIndexOutOfBoundsException("insufficient arguments for application");
			
			/*
			 * Store the name of the app, it will be forced into a
			 * new Flow and started up in Node.start()
			 */
			appVector.addElement(new AppConfig(command[2], command, 3));

		} else if (command[0].equals("extension")) {
			System.out.println("extension command ignored by ANTSR.");  // XXX
			// IGNORE.  Not used in Janos ANTS
			// FORMAT: extension <node addr> <class name> [<optional args> ... ]

		} else if (command[0].equals("manager")) {
			System.out.println("manager command ignored by ANTSR.");  // XXX
			// IGNORE.  Not used in Janos ANTS
			
		} else if (command[0].equals("connect")) {
			// IGNORE.  Its a legit command that is handled by makeroutes...
			
		} else {
			System.out.println("Ignored bad command " +command[0]+ "."); // XXX
		}
	}
	

	/**
	 */
	private final void parseCommands(Vector apps, String[] params)
	{
		// two required args
		String filename = params[0];
		String nodename = params[1];

		// Remainder go in a KeyArgs structure
		try
		{
			KeyArgs k = new KeyArgs(params, 2);
		}
		catch (InvalidKeyArgsException ikae)
		{
			FATALEXCEPTION(ikae, "ConfigurationManager args are bogus.");
		}
		
		/*
		 * Process the configuration file, creating entities
		 * as needed for each command.
		 */

		LineListTokenizer lt = null;
		try
		{
			lt = new LineListTokenizer(filename);
		}
		catch (FileNotFoundException fnfe)
		{
			panic(fnfe.getMessage());
		}
    
		String[] args;
		do
		{
			try
			{
				args = lt.list();
			}
			catch (LineListException lle)
			{
				FATALEXCEPTION(lle, "malformed line in " +filename+ ": " +lle.getMessage());
			}
			
			if (args == null)
				continue;

			if (args.length < 2)
			{
				// XXX do real i/o
				TRACE(Trace.ANTS_STARTUP, "Bad command/line " +args[0]+ ".");
				continue;
			}
		
			// Ignore commands that aren't for this node.
			if (! args[1].equals(nodename))
				continue;

			try
			{
				parseCommandLine(apps, args);			
			}
			catch (Exception e)
			{
				FATALEXCEPTION(e, "When parsing command " +args[0]+ ".");
			}
		}
		while (args != null);
	}
  
	/**
	 * Start a new configuration manager with the given args.
	 * Prints a usage message of insufficient arguments are given.
	 *
	 * @param args the arguments to pass to the ConfigurationManager constructor.
	 * @see #ConfigurationManager(java.lang.String[])
	 */
	public static void main(String[] args)
		throws Exception
	{
		// XXX uses System.out
		System.out.println("ConfigurationManager must be started as a Janos Flow.");
		System.exit(11);
	}

	protected BufferHandle newBufferHandle()
	{
		return new BufferHandle(flow);
	}

	protected void recycleBufferHandle(BufferHandle bh)
	{
	}

	protected void run(Flow flow,
			   ThreadPool threadPool,
			   String[] args,
			   CommSpaceHandle commSpaceHandle)
		throws Exception
	{
		DEBUGONLY(JANOS_CORE_PKG.Flow.currentFlow().setName("Primoridal Node Flow"));

		this.flow = flow; // for newBufferHandle();
		this.threadPool = threadPool; // for creating msgLoop thread
		
		if (args.length < 2) {
			System.out.println("usage: " 
					   +ConfigurationManager.class.getName()+
					   " <file> <node> [keyargs]");
			System.exit(1);
		}
    
		Vector apps = new Vector();
		parseCommands(apps, args);
		PrimordialNode.start(apps);
	}
}
