/*
 * Console.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_ANTS_ANTS_PKG;

import JANOS_CORE_PKG.Status;

import JANOS_IO_PKG.Log;

import JANOS_ANTS_UTILS_PKG.LineListTokenizer;
import JANOS_ANTS_UTILS_PKG.LineListException;

import java.net.ServerSocket;
import java.net.Socket;

import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.io.PrintWriter;

IMPORT_DEBUG
IMPORT_TRACE

PACKAGE class Console
	implements java.lang.Runnable // XXX
{
	private static final int MAX_SOCKET_BACKLOG = 2;

	private static final Status status = Status.getStatusObject();

	/**
	 * The socket this console listens on.
	 */
	private final ServerSocket socket;

	private final Node node;
	
	private PrintWriter out;

	private LineNumberReader in;

	private boolean done;

	/**
	 * Create a Console listener on the given port.
	 * Blows up if the port is already in use, throwing
	 * an IOException.
	 * <p>
	 * Connections will not be handled until the <code>accept()</code>
	 * method is executed.  You'd best supply a new thread
	 * to handle them, too.
	 *
	 * @see #accept()
	 * @param port the port to use; zero for any free port.
	 * @exception IOException throws whatever errors ServerSocket might toss your way.
	 */
	PACKAGE Console(Node node, int port)
		throws java.io.IOException // XXX 
	{
		this.node = node;

		// May throw IOException:
		this.socket = new ServerSocket(port, MAX_SOCKET_BACKLOG);
	}

	/**
	 * Return the port this console server listens on
	 */
	PACKAGE int getPort()
	{
		return socket.getLocalPort();
	}

	public String toString()
	{
		return "Console[" +socket.toString()+ "]";
	}

	public void run()
	{
		accept();
	}

	/**
	 * Accept and handle connections on this Console.
	 * <p>
	 * This call never returns.
	 */
	PACKAGE void accept()
	{
		while (true)
		{
			try
			{
				handleClient();
			}
			catch (java.io.IOException ioe)
			{
				TRACE(Trace.ANTS_CONSOLE, "console client exception: " +ioe.getMessage());
			}
		}
	}
	

	private void handleClient()
		throws java.io.IOException
	{
		Socket client;
		
		client = socket.accept();
		
		TRACE(Trace.ANTS_CONSOLE, "Accepted connection " +client);
		
		out = new PrintWriter(client.getOutputStream());
		in = new LineNumberReader(new InputStreamReader(client.getInputStream()));
		
		banner();
		
		while(!done)
		{
			// A prompt;
			out.print("> ");
			out.flush();

			String[] args;
			args = getCommand();
			
			if (args == null)
				continue;
			
			try
			{
				runCommand(args);
				out.flush();
			}
			catch (Throwable th)
			{
				node.log(Log.WARNING,
					 "Exception on console command " +args[0]+ ": " +th.getMessage());
				th.printStackTrace();
			}
		}
		in.close();
		out.close();
		client.close();
		TRACE(Trace.ANTS_CONSOLE, "Session complete.");

		// politely drop these on the floor..
		out = null;
		in = null;
	}

	private void banner()
	{
		out.println("Welcome to Janos ANTSR " +node);
		out.println("Type `exit' to exit.");
		out.flush();
	}

	private String[] getCommand()
		throws java.io.IOException
	{
		String line = in.readLine();
		String[] cmd;

		TRACE(Trace.ANTS_CONSOLE, "Got input line: " +line);

		line = line + "\n";
		LineListTokenizer lt = new LineListTokenizer(new StringReader(line));

		try
		{
			cmd = lt.list();
		}
		catch (LineListException lle)
		{
			cmd = null;
			out.println("Error parsing command: " +lle.getMessage());
		}

		// System.out.println("Parsed line into " +cmd.length+ " words.");
		// System.out.println("cmd[0] = " +((cmd!=null)?cmd[0]:"<cmd is null>"));

		return cmd;
	}

	private void usage()
	{
		out.println("Janos ANTSR Console help:");
		out.println("  exit / quit -- Close the console connection.");
		out.println("  shutdown    -- Shutdown this active node.");
		out.println("  help / ?    -- This message.");
		out.println("  filters     -- Packet filter map.");
		out.println("  flows       -- Show flows.");
		// flows -- show all the flows and their status
		// neighbors -- show all the immediate neighbors of this node
		// routes -- show all of the know destination hosts
		// stats -- packets sent, recv'd etc.
	}

	private final void dumpStrings(String[] s)
	{
		for (int i = 0; i < s.length; i++)
		{
			out.println(s[i]);
		}
	}
	
	private void runCommand(String[] args)
	{
		String cmd;
		
		if ((args == null) || (args[0] == null))
			return;

		cmd = args[0];

		if (cmd.equals("exit") || cmd.equals("quit"))
		{
			this.done = true;
		}
		else if (cmd.equals("shutdown"))
		{
			node.log(Log.WARNING,
				 "Shutdown issued from " +this.toString());
			node.shutdown();
		} 
		else if (cmd.equals("filters"))
		{
			String[] m = status.getPacketClassifierMap();
			out.println("Filter map:");
			dumpStrings(m);
		}
		else if (cmd.equals("flows"))
		{
			String[] m = status.getFlowMap();
			out.println("Flows:");
			dumpStrings(m);
		}
		else if (cmd.equals("help") || cmd.equals("?"))
		{
			usage();
		}
		else
		{
			out.println("Command " +cmd+ " is not recognized.");
			out.println("Type 'help' for a list of valid commands");
		}
		
	}
}

