/*
 * DynamicRouteEvents.Java
 * 
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package JANOS_ANTS_ROUTE_PKG;

import java.util.Vector;
import java.util.Enumeration;

IMPORT_TRACE
IMPORT_DEBUG

/**
 * Event queue for the dynamic route table.  Also contains
 * all of the timeout-related constants.
 */
PACKAGE class DynamicRouteEvents
{
	/**
	 * eventTime() rounds all timers to this granularity
	 * (in milliseconds).
	 *
	 * @see eventTime(long,long,long)
	 */
	private final static long RCRESOLUTION = 1000;

	// some RIP-like constants
	private final static int SECONDS = 1000;
	private final static int EXPIREINTERVAL = 60 * SECONDS;
	private final static int GCINTERVAL = 30 * SECONDS;
	private final static int UPDATEINTERVAL = 7 * SECONDS;
	private final static int UPDATESPREAD = 2 * SECONDS;
	private final static int TRIGGERINTERVAL = 2 * SECONDS;
	private final static int TRIGGERSPREAD = 1 * SECONDS;

	/**
	 * Store of timer events.  Each event is a DynamicRouteEvent.
	 * Not sorted or in any particular order (probably never
	 * long enough to justify).  There is exactly one
	 * event set to expire for any given time.
	 * <p>
	 * Used for expire, GC and periodic timers.
	 *
	 * @see #findTimer(long)
	 * @see #findFirstTimer()
	 * @see #setTimer(long,int)
	 * @see #clearTimer(long,int)
	 */
	private final Vector events;

	PACKAGE DynamicRouteEvents()
	{
		this.events = new Vector(16);
	}

	PACKAGE boolean isEmpty()
	{
		return events.isEmpty();
	}

	/**
	 * Compute a route event clock time.  All returned times
	 * are snapped to granularity of RCRESOLUTION.
	 *
	 * @param base Generally 'now'.
	 * @param add offset to add to base
	 * @param spread for randomizing the time a bit.
	 */
	PACKAGE static long eventTime(long base, long add, long spread) 
	{
		long time = base + add + (long)((double)spread*(Math.random()-0.5));
		assert(time > base);
		time += RCRESOLUTION - 1 ;
		time -= (time % RCRESOLUTION);
		return time;
	}
	
	/**
	 * Find a DynamicRouteEvent set to go off at "when".
	 * <p>
	 * Finds the first such event.
	 * <p>
	 * There is only ever be one (or zero) DynamicRouteEvent in the queue
	 * with a timer that matches when.
	 */
	PACKAGE DynamicRouteEvent findTimer(long when)
	{
		assert(when > 0);
		
		for (Enumeration e = events.elements(); e.hasMoreElements();)
		{
			DynamicRouteEvent r = (DynamicRouteEvent)e.nextElement();
			if (r.when == when)
			{
				return r;
			}
		}
		return null;
	}
	
	/**
	 * Find the event in the timer queue with the shortest
	 * timer.
	 */
	public DynamicRouteEvent findFirstTimer()
	{
		if (events.isEmpty())
			return null;
		
		DynamicRouteEvent low = (DynamicRouteEvent)events.firstElement();
		for (Enumeration e = events.elements(); e.hasMoreElements();)
		{
			DynamicRouteEvent r = (DynamicRouteEvent)e.nextElement();
			if (r.when < low.when) low = r;
		}
		
		return low;
	}
	
	/**
	 * Add a new timer to the event queue.
	 * <p>
	 * Reuses existing timers if one is set to go off at the same time.
	 */
	public DynamicRouteEvent setTimer(long when, int type)
	{
		DynamicRouteEvent re = findTimer(when);
		
		if (re == null)
		{
			re = new DynamicRouteEvent(when, type);
			events.addElement(re);
		}
		else
		{
			re.what |= type;
		}
		
		TRACE(Trace.ANTS_ROUTE_TIMER, "Adding timer, " + events.size() + " events pending");
		
		return re;
	}
	
	/**
	 * Remove the given DynamicRouteEvent from the list.
	 *
	 * @param r the event to remove
	 */
	public void clearTimer(DynamicRouteEvent r)
	{
		events.removeElement(r);
		
		TRACE(Trace.ANTS_ROUTE_TIMER, "Removing timer, " + events.size() + " events remain");
	}
	
	/**
	 * Add a expire timer to go off after a constant EXPIREINTERVAL.
	 * XXX can see RouteEntry.
	 *
	 * @see #EXPIREINTERVAL
	 */
	public void setExpireTimer(DynamicRouteEntry r, long now)
	{
		long when = eventTime(now, EXPIREINTERVAL, 0);
		r.setExpiration(when);
		DynamicRouteEvent re = setTimer(when, DynamicRouteEvent.EXPIRE);

		// Track the number of routes that will be expired at this time
		re.expireCount++;
	}
	
	/**
	 * Clear a expire timer from the event queue.  The event
	 * found at 'when' had best be a expire event with a non-zero
	 * expire count, thank you very much.
	 *
	 * @param when when the timer was scheduled to go off.
	 */
	public void clearExpireTimer(long when)
	{
		DynamicRouteEvent re = findTimer(when);

		// re must be a EXPIRE event
		if ((re==null) || ((re.what&DynamicRouteEvent.EXPIRE)==0) || (re.expireCount==0))
			panic("Bad reference count implementation for EXPIRE. re==" +re+ " @ " +when);
		
		// If the ref count goes to zero, then remove the event entirely.
		if (--re.expireCount == 0)
			re.what = re.what & (~DynamicRouteEvent.EXPIRE);

		// If the event isn't for anything anymore, expire the whole thing.
		if (re.what == DynamicRouteEvent.EMPTY)
			clearTimer(re);
	}
	
	/**
	 * Add a GC timer to go off a constant GCINTERVAL from now.
	 * Update the given route to 'expire' at that time.
	 * 
	 * XXX can see RouteEntry.
	 */
	public void setGCTimer(DynamicRouteEntry route, long now)
	{
		long when = eventTime(now, GCINTERVAL, 0);
		route.setExpiration(when);
		DynamicRouteEvent re = setTimer(when, DynamicRouteEvent.GC);
		re.gcCount++;
	}
	
	/**
	 * Clear a GC timer from the event queue.  There had
	 * best be a GC timer at moment when.
	 */
	public void clearGCTimer(long when)
	{
		DynamicRouteEvent re = findTimer(when);

		// GC timer had best exist
		if ((re == null) || ((re.what & DynamicRouteEvent.GC)==0) || (re.gcCount == 0))
			panic("Bad reference count implementation for GC");
		
		// If gcCount goes to zero, then drop it has a GC event
		if (--re.gcCount == 0)
			re.what = re.what & (~DynamicRouteEvent.GC);

		// If event has no purpose, remove it from queue.
		if (re.what == DynamicRouteEvent.EMPTY)
			clearTimer(re);
	}
	
	/**
	 * Add a periodic timer to go off at UPDATEINTERVAL +/- UPDATESPREAD
	 * millis after 'now'.
	 *
	 * @see #UPDATESPREAD
	 * @see #UPDATEINTERVAL
	 */
	public void setPeriodicTimer(long now)
	{
		long when = eventTime(now, UPDATEINTERVAL, UPDATESPREAD);
		setTimer(when, DynamicRouteEvent.PERIODIC);
	}
	
	/**
	 * Add a trigger timer to go off at TRIGGERINTERVAL +/- TRIGGERSPREAD
	 * millis after 'base'.
	 * <p>
	 * Sets the 'trigger' member to true.
	 *
	 * @see #TRIGGERINTERVAL
	 * @see #TRIGGERSPREAD
	 */
	public void setTriggerTimer(long base)
	{
		long when = eventTime(base, TRIGGERINTERVAL, TRIGGERSPREAD);
		setTimer(when, DynamicRouteEvent.TRIGGER);
	}
}
