/* 
 *  $Id: KeyArgs.Java,v 1.4 1999/10/18 20:23:11 tullmann Exp $
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package JANOS_ANTS_UTILS_PKG;

IMPORT_DEBUG

/**
 * A (key, argument) pair.  Used exclusively by KeyArgs.  Probably
 * should be an inner class.
 *
 * @author David Wetherall
 */
PACKAGE class KeyArg
{
	String key;
	String arg;

	/**
	 * Construct a new (key, argument) pair.
	 *
	 * @param k the key
	 * @param a the argument
	 */
	KeyArg(String key, String arg)
	{
		this.key = key;
		this.arg = arg;
	}
}


/**
 * Keyword argument manipulation utilities.  These are used
 * for handling command line arguments.  The key is any string
 * beginning with a "-"; the argument associated
 * with the key can be any string.
 * <p>
 * <b>WARNING: Do not try to pass an argument value of "-1" to these 
 * routines.  It will be recognized as a key, not an argument.</b>
 * (Any negative number will do that, of course.)
 * <p>
 * XXX fix the above, somehow.
 *
 * @author David Wetherall
 */
public class KeyArgs
{
	/**
	 * Array of KeyArgs, one for each option.
	 */
	private KeyArg[] args;

	/**
	 * The canonical 'null argument' KeyArg.
	 */
	private static final KeyArg empty = new KeyArg("", null);

	/**
	 * Return the number of key/arg pairs.
	 */
	public int length()
	{
		return args.length; 
	} 

	/**
	 * Return the i'th key.
	 */
	public String key(int i)
	{
		return args[i].key;
	}

	/**
	 * Return the i'th argument.
	 */
	public String arg(int i)
	{
		return args[i].arg;
	}

	/**
	 * Find the arg associated with the given key.  Returns
	 * 'null' if the key isn't found.
	 */
	public String get(String key)
	{
		for (int i = 0; i < args.length; i++)
			if (args[i].key.equals(key))
				return args[i].arg;
		return null;
	}

	/**
	 * Set the arg for key.  Overwrites existing
	 * value for key if one is set, otherwise
	 * adds a new pair to the list.
	 */
	public void set(String key, String arg) {
		for (int i = 0; i < args.length; i++)
		{
			if (args[i].key.equals(key))
			{
				args[i].arg = arg;
				return;
			}
		}
		
		KeyArg[] ret = new KeyArg[args.length + 1];  
	
		// XXX use System.arraycopy().
		for (int j = 0; j < args.length; j++)
			ret[j] = args[j];

		ret[args.length] = new KeyArg(key, arg);
		args = ret;
	}
  
	/**
	 * Strike the i'th key/arg pair from the list.
	 */
	public void strike(int i)
	{
		args[i] = empty;
	}

	/**
	 * Compact the list of key/arg pairs.  Removes all the
	 * "empty" ones.
	 */
	public void compact() {
		int len = 0;
		for (int i = 0; i < args.length; i++)
			if (args[i] != empty) len++;

		if (len == args.length)
			return;

		KeyArg[] ret = new KeyArg[len];
		int idx = 0;
		for (int i = 0; i < args.length; i++)
			if (args[i] != empty)
				ret[idx++] = args[i];

		args = ret;
	}
  
	/**
	 * Merge the given array of strings into the key/arg list.
	 */
	public void merge(String[] ka) 
		throws InvalidKeyArgsException
	{
		merge(new KeyArgs(ka));
	}

	/**
	 * Merge the given array of strings into the key/arg list.
	 * The first <code>skip</code> entries in the string array
	 * are skipped.
	 */
	public void merge(String[] ka, int skip) 
		throws InvalidKeyArgsException
	{
		merge(new KeyArgs(ka, skip));
	}

	/**
	 * Merge the given set of KeyArgs into the current list.
	 * Duplicates in <code>k</code> are thrown away.
	 */
	public void merge(KeyArgs k) {
		compact();

		if (k == null)
			return;
    
		k.compact();
	
		// Count the duplicates
		int len = args.length;
		for (int i = 0; i < k.args.length; i++) 
		{
			boolean match = false;
			for (int j = 0; j < args.length; j++)
				if (args[j].key.equals(k.args[i].key)) 
				{
					match = true;
					break;
				}
			if (!match)
				len++;
		}
    
		// XXX Use System.arraycopy()
		KeyArg[] ret = new KeyArg[len];  
		for (int j = 0; j < args.length; j++)
			ret[j] = args[j];
    
		int idx = args.length;
		for (int i = 0; i < k.args.length; i++) 
		{
			boolean match = false;
			for (int j = 0; j < args.length; j++)
				if (args[j].key.equals(k.args[i].key))
				{
					match = true;
					break;
				}

			if (!match)
				ret[idx++] = k.args[i];
		}

		args = ret;
	}

	/**
	 * Generate a nice, readable string representation of the key/arg
	 * pairs.
	 */
	public String toString()
	{
		compact();
		String ka = "";
		for (int i = 0; i < args.length; i++) {
			ka += "(\"" + args[i].key + "\", ";
			if (args[i].arg != null)
				ka += "\"" + args[i].arg + "\") ";
			else
				ka += "null) ";
		}
		return ka;
	}
  
	/**
	 * Generate an array of Strings, one for each
	 * Key/Arg pair.
	 */
	public String[] toStrings() {
		compact();
		int len = args.length;
		for (int i = 0; i < args.length; i++)
			if (args[i].arg != null)
				len++;

		String[] ret = new String[len];

		int idx = 0;
		for (int i = 0; i < args.length; i++) {
			ret[idx++] = args[i].key;
			if (args[i].arg != null)
				ret[idx++] = args[i].arg;
		}

		return ret;
	}

	public static void noMore(KeyArgs k)
	{
		// XXX for debugging:

		k.compact();
		if (k.length() > 0)
			panic("leftover key args: " +k);
	}

	private KeyArgs(KeyArg[] ka) {
		args = ka;
	}
  
	public KeyArgs(String[] ka)
		throws InvalidKeyArgsException
	{
		this(ka, 0);
	}

	public KeyArgs() 
		throws InvalidKeyArgsException
	{
		this(null, 0);
	}

	public KeyArgs(String[] ka, int skip)
		throws InvalidKeyArgsException
	{
		// handle the trivial case first
		if (ka == null || ka.length == 0 || ka.length <= skip) {
			args = new KeyArg[0];
			return;
		}
    
		// walk the list once, counting and checking syntax
		int l = 0;
		boolean ok = false;
		for (int i = skip; i < ka.length; i++) {
			if (ka[i] != null && ka[i].length() > 0 && ka[i].charAt(0) == '-') {
				l++;
				ok = true;
				continue;
			}
			if (!ok) {
				String ks = "";
				for (int j = 0; j < ka.length; j++)
					ks += ka[j] + " ";
				throw new InvalidKeyArgsException("invalid sequence: " + ks);
			}
			ok = false;
		}
    
		// now fill it in for real 
		args = new KeyArg[l];
		int idx = 0;
		for (int i = skip; i < ka.length; i++) {
			if (ka[i] != null && ka[i].length() > 0 &&  ka[i].charAt(0) == '-')
				args[idx++] = new KeyArg(ka[i], null);
			else
				args[idx-1].arg = ka[i];
		}
	}
}
