/*
 * FlowClassLoader.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

import java.util.Hashtable;

/**
 * Loads class files for a Flow.  There are two kinds of classes for
 * a FlowClassLoader: system classes and Flow classes.   The Flow
 * classes are specified when the FlowClassLoader is created by
 * the array of ClassSpecs.  Each ClassSpec specifies a name and
 * an array of bytes.  System classes are those loaded by the
 * parent ClassLoader.
 * <p>
 * A FlowClassLoader tries to find all loaded classes from the
 * already-loaded classes, then asks if its a system class,
 * and finally, checks the Flow-specific classes.
 */
PACKAGE final class FlowClassLoader 
	extends ClassLoader
{
	private Hashtable bytecodeTable = new Hashtable();

	/**
	 * Make a deep copy of the ClassSpec and push into
	 * the table of name to bytecode mappings.
	 * <p>
	 * FlowClassLoaders are only created when a new Flow() is
	 * created.  The ClassSpec elements have to be duplicated
	 * because the parameter is from the parent process's
	 * memory.
	 *
	 * @param classSpecs array of name and bytecode pairs.
	 */
	PACKAGE FlowClassLoader(ClassSpec[] classSpecs)
	{
		for(int i = 0; i < classSpecs.length; i++)
		{
			// XXX duplicate string?  Shared between processes?
			String s = new String(classSpecs[i].name);
			byte[] b = new byte[classSpecs[i].bytecode.length];
			System.arraycopy(classSpecs[i].bytecode, 0, b, 0,
				b.length);
			bytecodeTable.put(s, b);
		}
	}

	/**
	 * Add a new class to an existing flow.  This should
	 * be called from inside the flow that the class is
	 * being added to.
	 * Neither argument may be null.
	 */
	PACKAGE void addClass(String name, byte[] bytecode)
	{
		assert(name != null);
		assert(bytecode != null);
		bytecodeTable.put(name, bytecode);
	}

	public synchronized Class loadClass(String name, boolean resolve)
		throws ClassNotFoundException
	{
		Class c = findLoadedClass(name);
		if(c == null)
		{
			c = findSystemClass(name);
			if(c == null)
			{
				byte[] b = (byte[]) bytecodeTable.get(name);
				if(b == null)
					throw new ClassNotFoundException(name);
				c = defineClass(name, b, 0, b.length);
			}
		}
		if(resolve)
			resolveClass(c);
		return c;
	}
}
