/*
 * ThreadPool.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
package JANOS_CORE_PKG;

#ifdef BENCH
import JANOS_LIB_PKG.CycleTimer;
import JANOS_LIB_PKG.Semaphore;
#endif

/**
 * ThreadPool wrapper for Java.
 */
public final class ThreadPool
{
	/**
	 * Lock protecting the queue of WorkItems.
	 */
	private final Object workLock;

	/**
	 * WorkItems are added here.
	 */
	private WorkItem head;

	/**
	 * WorkItems are removed here.
	 */
	private WorkItem tail;

	private Flow flow;

#ifdef BENCH
	public static CycleTimer threadSchedTimer = new CycleTimer("ThreadPool thread schedule");
	public static CycleTimer threadDoneTimer = new CycleTimer("ThreadPool thread done");
	public static Semaphore serialLock = new Semaphore();
#endif //BENCH

	/**
	 * Create a new ThreadPool.
	 * <p>
	 * The variousSchedulerParameters and scheduler parameters
	 * hardly even qualify as hacks.
	 * <p>
	 * XXX This should take a ThreadPoolSpec as a parameter.
	 */
	public ThreadPool(Flow flow,
			  Object variousSchedulerParameters,
			  ThreadScheduler scheduler)
	{
		workLock = new Object();

		assert(flow != null);
		this.flow = flow;

		// XXX Start 8 worker threads.
		for (int i = 0; i < 8; i++)
#ifndef BENCH
			new Thread(this);
#else //BENCH
		{
			new Thread(this);
			serialLock.block();
		}
#endif //BENCH
	}

	/**
	 * Create an explicit thread and return that thread handle.
	 * This is a new thread which is not subject to whatever
	 * concurrency limit the ThreadPool has.  But, it will be
	 * scheduled by the ThreadPool's scheduler.
	 * 
	 * @allocations: This function will allocate a new Thread object.
	 * @param r the Runnable to start the target thread on.
	 */
	public Thread threadStart(java.lang.Runnable r)
	{
		return new Thread(this, r);
	}

	/**
	 * Add a task to the ThreadPool's task list.
	 * <p>
	 * XXX what happens if the same WorkItem is
	 * enqueued more than once?
	 */
	PACKAGE void addWork(WorkItem wi)
	{
		synchronized(workLock)
		{
			enqueueWorkItem(wi);
			signalWorkers();
		}
		BENCHONLY(serialLock.block());
		BENCHONLY(threadDoneTimer.end());
	}

	public Flow getFlow()
	{
		return this.flow;
	}

	PACKAGE boolean livePool()
	{
		return true;
	}

	PACKAGE void enqueueWorkItem(WorkItem wi)
	{
		assertLocked(workLock);
		assert(wi.next == null);
		assert(wi.prev == null);
		
		wi.prev = null;
		wi.next = head;

		if (head == null)
		{
			assert(tail == null);
			tail = wi;
		}
		else
		{
			assert(head != null);
			assert(tail != null);
			head.prev = wi;
		}

		head = wi;
	}

	PACKAGE WorkItem dequeueWorkItem()
	{
		assertLocked(workLock);

		if (head == null)
		{
			assert(tail == null);
			return null;
		}
		else
		{
			WorkItem wi;
			
			assert(head != null);
			assert(tail != null);

			wi = tail;
			tail = wi.prev;

			// if last item on list
			if (head == wi)
				head = null;
			else
				tail.next = null;

			DEBUGONLY(wi.next = null);
			DEBUGONLY(wi.prev = null);

			return wi;
		}
	}
	
	PACKAGE void signalWorkers()
	{
		assertLocked(workLock);
		BENCHONLY(threadSchedTimer.start());
		workLock.notify(); // XXX notifyAll() ???
	}

	/**
	 * Do the work on this ThreadPool.
	 * <p>
	 * This function never returns.
	 */
	PACKAGE void doWork()
	{
		BENCHONLY(serialLock.signal());
		WorkItem wi = null;
		while(livePool())
		{
			synchronized(workLock)
			{
				try
				{
#ifndef BENCH
					wi = dequeueWorkItem();
					if (wi == null)
						workLock.wait();
#else //BENCH
					workLock.wait();
					wi = dequeueWorkItem();
					assert(wi != null);
#endif //BENCH
				}
				catch (java.lang.InterruptedException ie)
				{
					// from the wait().  Just ignore it.
				}
			}

			// Do the item outside the context of the workLock!
			if (wi != null)
			{
				BENCHONLY(threadSchedTimer.end());

				// do the item
				wi.doItem();

				// Cannot do anything with the item at this point,
				// it might already be scheduled for re-use.
				wi = null;

				// XXX If an exception is throw, all sorts
				// of problems are caused!  What to do
				// with the exceptions???


				BENCHONLY(threadDoneTimer.start());
				BENCHONLY(serialLock.signal());
			}
		}
	}
}
