# genmk.pl
#
# Copyright (c) 1999 University of Utah and the Flux Group.
# All rights reserved.
#
# This file is part of Janos.
# 
# Janos is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
# 
# Janos is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, in the file
# COPYING.  If that file is not present, write to the Free
# Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
# 02111-1307, USA
# 
# The University of Utah and the Flux Group requests users to
# return to flux-dist@cs.utah.edu any improvements that they make.
# 
# Written by Patrick Tullmann <tullmann@cs.utah.edu>
#

#
# Generate a makefile for testing an application with
# various command line arguments, and "failure modes."
#
# Expects two names on the command line, the first is the name of the
# configuration file describing the test.  The second is the name of
# the makefile to generate from that configuration.  This filename
# should end in '.mk'.  This target filename is also used to construct
# the basename of the tests.  So, a target filename of 'crusher.mk'
# creates a file called 'crusher.mk' and all of the targets in
# this makefile are prefixed with 'crusher-'.
#
# Configuration files may contain comments (just like this one).
# The following directives are recognized:
#  NAME=<name>  declares a new test with name <name>.
#  DIFFOPT=various arguments to be passed to the test differentiator. 
#    (see runtest.pl for the supported arguments)
#  CLASS=<name> the name of the classfile to pass to Java.
#  TIMEOUT=<val> the longest the test engine should wait for a test to
#    to finish.
#  VMARGS=<args> the arguments to pass to the VM (before the class)
#  ARGS=<args> the arguments to pass to the CLASS.
#
# The generated makefile contains the following targets: (basename is the name 
# of the makefile minus '.mk'.)
#  <basename>-setup: to setup all targets (compile and generate "reference" output)
#  <basename>-run: to run all targets (and compare output to the reference output)
#

$doClassDependency = 0;

$conf = shift;
$target = shift;

$basename = $target;
$basename =~ s/\.mk$//;

open(TARGET, ">$target") || die "Cannot open $target: $!\n";

print TARGET "### THIS FILE WAS AUTOMATICALLY GENERATED FROM $conf\n";
print TARGET "### To rengerate run 'genmk $conf $target'\n";
print TARGET "\n";

$allSetupTargets = '';
$allRunTargets = '';

$testname = '';

$script = '$(JAVA)';  ### default is to run the JVM
$vmargs = '';
$args = '';
$diffArgs = '';
$timeout = '';  ### seconds
$classname = $basename;

open(CONF, "$conf") || die "Cannot open $conf: $!\n";

$wroteNothing = 1;

while (<CONF>) {
    chop;
    s/\#.*//;			# throw away comments

    if (/NAME=(.*)/) {
	## Write out previous test
	&writeRules() if ($testname ne '');

	## Setup the new test (re-set basic variables)
	$testname = "${basename}-$1";

	$args = '';
	$vmargs = '';
	$diffArgs = '';
	$timeout = '';
	$classname = $basename;
    }
    
    if (/DIFFOPTS?=(.*)/) {
	$diffArgs = $1;
	$diffArgs =~ s/\s*//;
    }

    if (/SCRIPT=(.*)/) {
        $script = $1;
	$classname = ''; ### XXX not really the right thing to do.
    }
    
    if (/CLASS=(.*)/) {
	$classname = $1;
    }

    if (/TIMEOUT=(.*)/) {
        ## seconds
	$timeout = $1;
    }

    if (/VMARGS=(.*)/) {
	$vmargs = $1;
    }
    
    if (/ARGS=(.*)/) {
	$args = $1;
    }
}

if ($wroteNothing != 0) {
    $testname="${basename}-default";
}

## write out the last rule
&writeRules();

## Write trailer
print TARGET "### Make the FORCE target do nothing\n";
print TARGET "FORCE:\n\n";

print TARGET "### Make all of the fake targets in this file phony.\n";
print TARGET ".PHONY: FORCE $allSetupTargets $allRunTargets $basename-setup $basename-run\n\n";

print TARGET "### Define entrypoints that will run all the tests in this makefile\n";
print TARGET "$basename-setup: $allSetupTargets\n\n";
print TARGET "$basename-run: $allRunTargets\n";

close(TARGET);
close(CONF); 

# END

### ---------------------------------------------

sub writeRules() {
    local($ref) = "\$(REFOUTPUT_DIR)/${testname}.\$(REFOUTPUT_SUFFIX)";
    local($command) = "$script $vmargs $classname $args";
    local($diffOpt) = '';
    local($timeoutOpt) = '';
    local($doQuiet) = '@';

    ## Cleanup command
    $command = join(' ', split(' ', $command));
    $command = "\'${command}\'";

    $diffOpt = "--diffopt=$diffArgs" if ($diffArgs ne '');
    $timeoutOpt = "--timeout=$timeout" if ($timeout ne '');

    $wroteNothing = 0;

    $allSetupTargets .= " $testname-setup";
    $allRunTargets .= " $testname-run";

    print TARGET "TESTCASES+= $testname\n\n";
    print TARGET "### run and setup rules for $testname (args='$args')\n";
    ## Setup rule
    if ($doClassDependency) {
	print TARGET "$testname-setup: $classname.class\n";
    } else {
	print TARGET "$testname-setup: FORCE\n";
    }
    print TARGET "\t${doQuiet}rm -f $ref \n";
    print TARGET "\t${doQuiet}echo \"   Generating reference for ${command}\";\n";
    print TARGET "\t${doQuiet}\$(RUNTEST) $diffOpt $timeoutOpt --setup $command $ref \n";
    print TARGET "\n";

    ## Run rule
    if ($doClassDependency) {
	print TARGET "$testname-run: $classname.class $ref\n";
    } else {
	print TARGET "$testname-run: $ref\n";
    }
    print TARGET "\t${doQuiet}echo \"   Running $command\";\n";
    print TARGET "\t${doQuiet}\$(RUNTEST) \$(RUNTESTDEBUG) $diffOpt $timeoutOpt --logfile=\$(LOGFILE) --compare $command $ref \n";
    print TARGET "\n";
}


# eof
