#
# JavaClass.pm
#
# Copyright (c) 1999 University of Utah and the Flux Group.
# All rights reserved.
#
# This file is part of Janos.
# 
# Janos is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
# 
# Janos is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, in the file
# COPYING.  If that file is not present, write to the Free
# Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
# 02111-1307, USA
# 
# The University of Utah and the Flux Group requests users to
# return to flux-dist@cs.utah.edu any improvements that they make.
# 
# Written by Patrick Tullmann <tullmann@cs.utah.edu>
#

#
# Functions for reading in and writing out a Java .class file.
# Also does a bit of consistency checking of the file.
#
# The only really nasty thing I've done (because of poor perl skils more
# than anything else) is to make the %class a local() in a number of
# places so that the check routines can see it.
#
# Class structure:  Generally references to hashes. Tables are implemented as arrays.
#
# TODO: 
#	make a &checkClass() function.
#	change a lot of 'local's to 'my's. (not local(%class), though)
#	Make CLASSIN and CLASSOUT parameters to read/write functions.
#	POD documentation
#	Cannot handle modifying float values.  I can read and decode, but don't
#	have the math to convert back to a binary format (both floats and doubles).

package JavaClass;

###
### Define constants for Java Classes
###

*classMagic = \0xcafebabe;	# The magic header every .class file starts with

## The magic identifiers for entries in the .class Constant Table.
*CONSTANT_Class = \7;
*CONSTANT_FieldRef = \9;
*CONSTANT_MethodRef = \10;
*CONSTANT_InterfaceMethodRef = \11;
*CONSTANT_String = \8;
*CONSTANT_Integer = \3;
*CONSTANT_Float = \4;
*CONSTANT_Long = \5;
*CONSTANT_Double = \6;
*CONSTANT_NameAndType = \12;
*CONSTANT_Utf8 = \1;

## String names associated with each type of Constant Table entry.
%CONSTANTNames = (
    $CONSTANT_Class => "Class",
    $CONSTANT_FieldRef => "Field",
    $CONSTANT_MethodRef => "Method",
    $CONSTANT_InterfaceMethodRef => "Inteface Method",
    $CONSTANT_String => "String",
    $CONSTANT_Float => "Float",
    $CONSTANT_Integer => "Integer",
    $CONSTANT_Double => "Double",
    $CONSTANT_Long => "Long",
    $CONSTANT_NameAndType => "Name&Type",
    $CONSTANT_Utf8 => "Utf8"
    );

@instructionLength = (
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        2, 3, 2, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 
        3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 0, 0, 1, 1, 1, 1, 
        1, 1, 3, 3, 3, 3, 3, 3, 3, 5, 1, 3, 2, 3, 1, 1, 
        3, 3, 1, 1, 1, 4, 3, 3, 5, 5, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 
);



## String names for the shorthand used in signatures
$sig{'V'} = 'void';
$sig{'I'} = 'int';
$sig{'J'} = 'long';
$sig{'Z'} = 'boolean';
$sig{'F'} = 'float';
$sig{'D'} = 'double';
$sig{'B'} = 'byte';
$sig{'S'} = 'short';
$sig{'C'} = 'char';

## Access control flags for classes, methods and fields.
*ACC_PUBLIC    = \0x0001;
*ACC_PRIVATE   = \0x0002;
*ACC_PROTECTED = \0x0004;
*ACC_STATIC    = \0x0008;
*ACC_FINAL     = \0x0010;
*ACC_SUPER     = \0x0020;
*ACC_VOLATILE  = \0x0040;
*ACC_TRANSIENT = \0x0080;
*ACC_INTERFACE = \0x0200;
*ACC_ABSTRACT  = \0x0400;
*ACC_NATIVE    = \0x0100;

*ACC_UNKNOWN   = \0xF800;

###
###  Global variables
###

# Control the verbosity of &printClass()
$detailedFields = 0;
$detailedMethods = 0;
$showMethodCode = 1;

###
### Conversion functions
###

## parseJavaSig() takes a single argument, a single Java-internal
## method signature and returns a list ($package, $return, $class,
## $method, @args) where the items have been converted to a more
## source-like format (e.g., english).
sub parseJavaSig() {
  ## Parameters
  my $jsig = shift;

  ## Local variables
  my $class = '';
  my $package = '';
  my $method = '';
  my @args = ();
  my $ret = '';

  ## Temporaries
  my $depth = 0;
  my $repct = 0;
  my $arg = '';

  ### First is the class (all chars until a ".")
  $jsig =~ s/^([^.]*).//;
  $class = $1;
  $class =~ s,/,.,g; # / -> .

  # Peel the package name out of the class name (everything before last ".")
  if ($class =~ m/(.*)\.[^\.]*$/) {
    $package = $1;
  }

  ### Second comes the method name (all chars until a left paren)
  $jsig =~ s/^([^\(]*)\(//;
  $method = $1;

  ### Now the arguments
 SIGPARSE:
  while(1) {
    $repct = $jsig =~ s/^(I|J|Z|F|D|B|S|C|L|\[|\))//;  ## No V types
    die "badly formed signature at $jsig" if ($repct == 0);
    $arg = $1;
    
    ## Stop if we hit the end paren
    last SIGPARSE if $arg eq "\)";
    
    if ($arg eq '[') {
      $depth++;
      # continue parsing array type... 
      next SIGPARSE;
    } elsif ($arg eq 'L') {
      $jsig =~ s/^([^;]*);//;
      $arg = $1;
      $arg =~ s,/,.,g;
    } else {
      ## convert single-char identifier to english
      $arg = $sig{$arg};
    }
    
    ## If we hit an array, tack the array depth on the end
    if ($depth > 0) {
      $arg = $arg . "[]" x $depth;
      $depth = 0;
    }

    # Put the arg at the end of the list of args
    push (@args, $arg)
  }

  ### Last is the return type
  $depth = 0;
  $repct = $jsig =~ s/^(I|J|Z|F|D|B|S|C|L|V|\[)//; ## Adds V over argument types
  die "badly formed return type: \'$jsig\'" if ($repct == 0);
  $ret = $1;

  # If its an array, eat the [ and re-set $ret
  if ($ret eq '[') {
    $depth = 1;
    while ($jsig =~ s/\[//) {
      $depth++;
    }
    $jsig =~ s/^(I|J|Z|F|D|B|S|C|L)//; ## No [ or V
    die "badly formed return type: \'$jsig\'" if ($repct == 0);
    $ret = $1;
  } 

  if ($ret eq 'L') {
    $jsig =~ s/^([^;]*);//;
    $ret = $1;
    $ret =~ s,/,.,g;
  } else {
    ## Convert single char identifier to english
    $ret = $sig{"$ret"};
  }

  # Tack the array brackets on
  if ($depth > 0) {
    $ret = $ret . "[]" x $depth;
  }
    
  ### Return the info in an easy-to-use list
  return ($package, $ret, $class, $method, @args);
}

###
### Print functions
###

sub printClass {
    my $r_cl = shift;		
    my %class = %{$r_cl};

    my $flStr = &ACCFlagsToString($class{accessFlags});
    print "$flStr\n";

    &printConstantPool($r_cl);

    ## Print 'this_class'
    my $thisClName = %{$class{constantPool}[$class{thisClass}]}->{nameIndex};
    $thisClName = %{$class{constantPool}[$thisClName]}->{val};
    print "this_class @ $class{thisClass} ($thisClName)\n";

    ## Print 'super_class'
    if ($class{superClass} != 0) {
      my $superClName = %{$class{constantPool}[$class{superClass}]}->{nameIndex};
      $superClName = %{$class{constantPool}[$superClName]}->{val};
      print "super_class @ $class{superClass} ($superClName)\n";
    } else {
      print "No super class\n";
    }

    ## Print direct super interfaces
    &printInterfaces($r_cl);

    ## Print fields
    &printFields($r_cl);

    ## Print methods
    &printMethods($r_cl);

    ## Print attributes
    &printAttributes("", $r_cl, $class{attributes});
}

sub printMethods {
    my $r_cl = shift;
    local(%class) = %{$r_cl};

    if ($class{methodCt} == 0) {
	print "No fields.\n";
    } else {
	$i = 0;
	print "Methods:\n";
	while ($i < $class{methodCt}) {
	    my %method = %{$class{methods}[$i]};
    
	    my $accflags = ACCFlagsToString($method{accessFlags});
	    my $name = $class{constantPool}[$method{nameIndex}]->{val};
	    my $desc = $class{constantPool}[$method{descriptorIndex}]->{val};

	    if ($detailedMethods) {
		print ("\t$i: ");
		print (".accessFlags=$accflags; ");
		print (".name @ $method{nameIndex} ($name); ");
		print (".descriptor @ $method{descriptorIndex} ($desc); ");
		print (".attrCt = $method{attributesCt};\n");
		&printAttributes("\t\t", \%class, $method{attributes});
	    } else {
		print ("\t$accflags $name $desc\n");
	    }

	    if ($showMethodCode) {
	      &printCodeAttr("\t\t", \%class, $method{attributes});
	    }

	} continue {
	    $i++;
	}
    }
}

sub printFields {
    my $r_cl = shift;
    local(%class) = %{$r_cl};

    if ($class{fieldCt} == 0) {
	print "No fields.\n";
    } else {
	$i = 0;
	print "Fields:\n";
	while ($i < $class{fieldCt}) {
	    my %field = %{$class{fields}[$i]};
    
	    my $accflags = ACCFlagsToString($field{accessFlags});
	    my $name = $class{constantPool}[$field{nameIndex}]->{val};
	    my $desc = $class{constantPool}[$field{descriptorIndex}]->{val};

	    if ($detailedFields) {
		print ("\t$i: ");
		print (".accessFlags=$accflags; ");
		print (".name @ $field{nameIndex} ($name); ");
		print (".descriptor @ $field{descriptorIndex} ($desc); ");
		print (".attrCt = $field{attributesCt};\n");
		&printAttributes("\t\t", \%class, $field{attributes});
	    } else {
		print ("\t$accflags $desc $name\n");
	    }
	} continue {
	    $i++;
	}
    }
}

sub printInterfaces {
    my $r_cl = shift;
    local(%class) = %{$r_cl};

    if ($class{interfaceCt} == 0) {
	print "No interfaces.\n";
    } else {
	my $i = 0; 
	print "Interfaces:\n";
	while ($i < $class{interfaceCt}) {
	    my $iClass = $class{interfaces}[$i];
	    my %iClassConst = %{$class{constantPool}[$iClass]};
	    my $iClassName = $iClassConst{nameIndex};
	    my $interfaceName = %{$class{constantPool}[$iClassName]}->{val};
	    print "\t$i] @ $iClass ($interfaceName)\n";
	} continue {
	    $i++;
	}
    }
}

sub ACCFlagsToString {
    my $flags = shift;
    my @flags = ();

    push(@flags, "public") if $flags & $ACC_PUBLIC;
    push(@flags, "private") if $flags & $ACC_PRIVATE;
    push(@flags, "protected") if $flags & $ACC_PROTECTED;
    push(@flags, "static") if $flags & $ACC_STATIC;
    push(@flags, "final") if $flags & $ACC_FINAL;
    push(@flags, "super") if $flags & $ACC_SUPER;
    push(@flags, "volatile") if $flags & $ACC_VOLATILE;
    push(@flags, "transient") if $flags & $ACC_TRANSIENT;
    push(@flags, "interface") if $flags & $ACC_INTERFACE;
    push(@flags, "abstract") if $flags & $ACC_ABSTRACT;
    push(@flags, "native") if $flags & $ACC_NATIVE;
    push(@flags, "UNKNOWN") if $flags & $ACC_UNKNOWN;

    return join(',', @flags);
}

sub printConstantPool {
    my $r_cl = shift;
    local(%class) = %{$r_cl};		# cvt the class reference to the class hash

    print("Constant Pool Entries: $class{constantPoolCt}\n");

    $i = 1;
    while ($i < $class{constantPoolCt}) {
	my %cpEntry = %{$class{constantPool}[$i]};

	print "$i] $CONSTANTNames{$cpEntry{tag}}: ";

    	if ($cpEntry{tag} eq $CONSTANT_Class) {
	    my $ni = $cpEntry{nameIndex};

    	    &checkIndex($ni, "Name", $CONSTANT_Utf8);
    
	    my $nm = $class{constantPool}[$ni]->{val};

	    print (".name @ $ni ($nm);");

    	} elsif (($cpEntry{tag} eq $CONSTANT_FieldRef)
		 || ($cpEntry{tag} eq $CONSTANT_MethodRef)
		 || ($cpEntry{tag} eq $CONSTANT_InterfaceMethodRef)) {
    	    &checkIndex($cpEntry{classIndex}, "Class", $CONSTANT_Class);
    	    &checkIndex($cpEntry{nameTypeIndex}, "Name & Type", $CONSTANT_NameAndType);

	    print (".class @ $cpEntry{classIndex}; .name&type @ $cpEntry{nameTypeIndex};");
    	} elsif ($cpEntry{tag} eq $CONSTANT_String) {
	    my $si = $cpEntry{stringIndex};
    	    &checkIndex($si, "String", $CONSTANT_Utf8);
	    
	    my $str = $class{constantPool}[$si]->{val};

	    print (".string @ $cpEntry{stringIndex} ($str);");
    	} elsif ($cpEntry{tag} eq $CONSTANT_NameAndType) {
	    my $ni = $cpEntry{nameIndex};
	    my $di = $cpEntry{descriptorIndex};

    	    &checkIndex($ni, "Name", $CONSTANT_Utf8);
    	    &checkIndex($di, "Descriptor", $CONSTANT_Utf8);

	    my $nstr = $class{constantPool}[$ni]->{val};
	    my $dstr = $class{constantPool}[$di]->{val};

	    print (".name @ $ni ($nstr); .descriptor @ $di ($dstr); ");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Integer) {
	    print (".value = $cpEntry{val}");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Utf8) {
	    print (".length=$cpEntry{len}; ");
	    print (".val=$cpEntry{val}; ");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Float) {
	    print (".val=$cpEntry{strVal}; ");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Double) {
	    print (".val=$cpEntry{strVal}; ");
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} elsif ($cpEntry{tag} eq $CONSTANT_Long) {
	    print (".val=$cpEntry{strVal}; ");
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} else {
    	    &fatal("Unknown Constant type $cpEntry{tag}!\n");
    	}

	print ("\n");
	
	$i++;
    }
}

## print the Code attribute.
## arguments are (string prefix, ref to the class, attributes block ref)
sub printCodeAttr {
  my ($prefix, $r_class, $r_attrs) = @_;

  return if (!defined($r_attrs));

  foreach $r_attr (@{$r_attrs}) {
    my $attrName = $class{constantPool}[$r_attr->{nameIndex}]->{val};

    if ($attrName eq "Code") {
      &printBytecodes("$prefix  ", $r_attr->{code}, $r_attr->{codeLength});

      printf "${prefix}maxStack: $r_attr->{maxStack}\n";
      printf "${prefix}maxLocals: $r_attr->{maxLocals}\n";
      # printf "${prefix}codeLength: $r_attr->{codeLength}\n";
    }
  }
}

sub printBytecodes() {
  my ($prefix, $code, $codelen) = @_;
  my $offset = 0;
  
  while ($offset < $codelen) {
    my $instr = &grab_u1($offset, $code);
    my $ilen = $instructionLength[$instr];
    my $iname = $BytecodeNames{$instr};
    #print "$BytecodeNames{$instr} ($instr) ($ilen bytes)\n", 
    printf "${prefix}%4d] %s\n", $offset, $iname;

    $offset += $ilen;
  }
}


sub grab_u1 {
  my ($offset, $string) = @_;
  my $val = 0;
  ($val) = unpack("C", substr($string, $offset, 1));
  return $val;
}

sub grab_u4 {
  my ($offset, $string) = @_;
  my $val = 0;
  ($val) = unpack("N", substr($string, $offset, 4));
  return $val;
}

sub grab_u2 {
  my ($offset, $string) = @_;
  my $val = 0;
  ($val) = unpack("n", substr($string, $offset, 2));
  return $val;
}


sub printAttributes {
    my ($prefix, $r_class, $r_attrs) = @_;

    return if (!defined($r_attrs));

    my $i = 0;
    my %class = %{$r_class};
    
    print ("${prefix}Attributes:\n");
    foreach $r_attr (@{$r_attrs}) {

	my $name = $class{constantPool}[$r_attr->{nameIndex}]->{val};
	print ("${prefix}\t.name=$name; ");
	print (".length=" . $r_attr->{len} . ";");

	if ($name eq 'SourceFile') {
	  if ($r_attr->{len} != 2) {
	    print ("!Badly formed SourceFile Attribute, must be 2!");
	  } else {
	    my ($high, $low) = unpack("CC", $r_attr->{rawattr});
	    my $idx = ($high * 256) + $low;
	    my $name = $class{constantPool}[$idx]->{val};
	    print (" @ " . $idx . " (\"" . $name . "\")");
	  }
	}

	print ("\n");
    }
}

###
### Read Class function
###

sub readClass {
    my $classFile = shift;
    local(%class) = (());

    open(CLASSIN, $classFile) 
	|| open(CLASSIN, "${classFile}.class")
	    || die ("Cannot open $classFile for reading");
    
    ###
    ### Header Magic
    ###
    	
    $class{magic} = read_u4();
    if ($class{magic} != $classMagic) {
    	fatal("Bad class magic '$class{magic}' --expected '$classMagic'.  $classFile is probably not a Java class file.");
    }
    
    ## Read in the major and minor version numbers
    $class{minorVersion} = &read_u2();
    $class{majorVersion} = &read_u2();
    
    print("Version: $class{majorVersion}.$class{minorVersion}  (expected 45.3)\n") 
    	if ($class{minorVersion} ne 3) || ($class{majorVersion} ne 45);
    
    ###
    ### Constant Pool
    ###
    $class{constantPoolCt} = &read_u2();
    $class{constantPool} = [];
    
    $i = 1; # constant pool actually starts with entry 1...
    while ($i < $class{constantPoolCt}) {
    	my %cpEntry;
    	$cpEntry{tag} = &read_u1();
    
    	if ($cpEntry{tag} eq $CONSTANT_Class) {
    	    $cpEntry{nameIndex} = &read_u2();
    
    	    &checkIndex($cpEntry{nameIndex}, "Name");
    	} elsif ($cpEntry{tag} eq $CONSTANT_FieldRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    	    
    	    &checkIndex($cpEntry{classIndex}, "Class");
    	    &checkIndex($cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_MethodRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    
    	    &checkIndex($cpEntry{classIndex}, "Class");
    	    &checkIndex($cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_InterfaceMethodRef) {
    	    $cpEntry{classIndex} = &read_u2();
    	    $cpEntry{nameTypeIndex} = &read_u2();
    
    	    &checkIndex($cpEntry{classIndex}, "Class");
    	    &checkIndex($cpEntry{nameTypeIndex}, "Name & Type");
    	} elsif ($cpEntry{tag} eq $CONSTANT_String) {
    	    $cpEntry{stringIndex} = &read_u2();
    
    	    &checkIndex($cpEntry{stringIndex}, "String");
    	} elsif ($cpEntry{tag} eq $CONSTANT_NameAndType) {
    	    $cpEntry{nameIndex} = &read_u2();
    	    $cpEntry{descriptorIndex} = &read_u2();
    
    	    &checkIndex($cpEntry{nameIndex}, "Name");
    	    &checkIndex($cpEntry{descriptorIndex}, "Descriptor");
    	} elsif ($cpEntry{tag} eq $CONSTANT_Integer) {
    	    $cpEntry{val} = &read_u4();
    	} elsif ($cpEntry{tag} eq $CONSTANT_Utf8) {
    	    $cpEntry{len} = &read_u2();
    	    $cpEntry{val} = &read_utf8($cpEntry{len});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Float) {
	    $cpEntry{val} = &read_u4();
	    $cpEntry{strVal} = &parse_float($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Double) {
	    $cpEntry{val} = &read_u8();
	    $cpEntry{strVal} = &parse_double($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Long) {
	    $cpEntry{val} = &read_u8();
	    $cpEntry{strVal} = &parse_long($cpEntry{val});
    	} else {
    	    &fatal("Unknown Constant type $cpEntry{tag}!\n");
    	}
    
    	$class{constantPool}[$i] = \%cpEntry;

	## Ick.  8-byte entries take two constant pool entries
	$i++ if (($cpEntry{tag} == $CONSTANT_Long)
		 || ($cpEntry{tag} == $CONSTANT_Double));
    } continue {
    	$i++;
    }
    
    ###
    ### Misc. Class Info
    ###
    
    $class{accessFlags} = &read_u2();
    
    $class{thisClass} = &read_u2();
    &checkIndex($class{thisClass}, "this_class", $CONSTANT_Class);
    
    $class{superClass} = &read_u2();
    if ($class{superClass} != 0) {
      &checkIndex($class{superClass}, "super_class", $CONSTANT_Class);
    } else {
      print ("Warning: class has no super class.  Must be java.lang.Object\n"); 
    }
    
    ###
    ### Direct super-interfaces
    ###
    $class{interfaceCt} = &read_u2();
    $class{interfaces} = [];
    
    $i = 0;
    while ($i < $class{interfaceCt}) {
    	$class{interfaces}[$i] = &read_u2();
    
    	&checkIndex($class{interfaces}[$i], "Interface \#$i", $CONSTANT_Class);	
    } continue {
    	$i++;
    }
    
    ###
    ### Fields
    ###
    $class{fieldCt} = &read_u2();
    $class{fields} = [];
    
    $i = 0;
    while ($i < $class{fieldCt}) {
    	my %field;
    	$field{accessFlags} = &read_u2();
    	$field{nameIndex} = &read_u2();
    	$field{descriptorIndex} = &read_u2();
    	$field{attributesCt} = &read_u2();
    	$field{attributes} = &readAttributes($class, $field{attributesCt});
    
    	&checkIndex($field{nameIndex}, "Field Name", $CONSTANT_Utf8);
    	&checkIndex($field{descriptorIndex}, "Field Descriptor", $CONSTANT_Utf8);
    
    	$class{fields}[$i] = \%field;
    } continue {
    	$i++;
    }
    
    ###
    ### Methods
    ###
    $class{methodCt} = &read_u2();
    $class{methods} = [];
    
    $i = 0;
    while ($i < $class{methodCt}) {
    	my %method;
    
    	$method{accessFlags} = &read_u2();
    	$method{nameIndex} = &read_u2();
    	$method{descriptorIndex} = &read_u2();
    	$method{attributesCt} = &read_u2();
    	$method{attributes} = &readAttributes($class, $method{attributesCt});
    
    	&checkIndex($method{nameIndex}, "Method Name", $CONSTANT_Utf8);
    	&checkIndex($method{descriptorIndex}, "Method Descriptor", $CONSTANT_Utf8);
    
    	$class{methods}[$i] = \%method;
    } continue {
    	$i++;
    }
    
    ###
    ### Class attributes
    ###
    $class{attributesCt} = &read_u2();
    $class{attributes} = &readAttributes($class, $class{attributesCt});

    ###
    ### End of .class file
    ###

    return \%class;
}

###
### Write Class function
###

sub writeClass {
    my $r_class = shift;
    my $classFile = shift;
    local(%class) = %{$r_class};

    if ($classFile =~ /\.class$/) {
	open(CLASSOUT, ">$classFile") 
	    || die ("Cannot open $classFile for writing");
    } else {
	open(CLASSOUT, ">$classFile.class") 
	    || die ("Cannot open $classFile.class for writing");
    }

    ###
    ### Header Magic
    ###

    if ($class{magic} != $classMagic) {
    	fatal("Bad class magic '$class{magic}' --expected '$classMagic'.  Not writing class file.");
    }
    	
    &write_u4($class{magic});
    
    ## Write major/minor version numbers
    &write_u2($class{minorVersion});
    &write_u2($class{majorVersion});
    
    ###
    ### Constant Pool
    ###
    &write_u2($class{constantPoolCt});

    $i = 1; # constant pool actually starts with entry 1...
    while ($i < $class{constantPoolCt}) {
    	my %cpEntry = %{$class{constantPool}[$i]};
    	&write_u1($cpEntry{tag});
    
    	if ($cpEntry{tag} eq $CONSTANT_Class) {
    	    &checkIndex($cpEntry{nameIndex}, "Name", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{nameIndex});
    	} elsif (($cpEntry{tag} eq $CONSTANT_FieldRef)
		 || ($cpEntry{tag} eq $CONSTANT_MethodRef)
		 || ($cpEntry{tag} eq $CONSTANT_InterfaceMethodRef)) {
    	    &checkIndex($cpEntry{classIndex}, "Class", $CONSTANT_Class);
    	    &checkIndex($cpEntry{nameTypeIndex}, "Name & Type", $CONSTANT_NameAndType);

    	    &write_u2($cpEntry{classIndex});
    	    &write_u2($cpEntry{nameTypeIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_String) {
    	    &checkIndex($cpEntry{stringIndex}, "String", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{stringIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_NameAndType) {
    	    &checkIndex($cpEntry{nameIndex}, "Name", $CONSTANT_Utf8);
    	    &checkIndex($cpEntry{descriptorIndex}, "Descriptor", $CONSTANT_Utf8);

    	    &write_u2($cpEntry{nameIndex});
    	    &write_u2($cpEntry{descriptorIndex});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Integer) {
    	    &write_u4($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Utf8) {
    	    &write_u2($cpEntry{len});
    	    &write_utf8($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Float) {
	    &write_u4($cpEntry{val});
    	} elsif ($cpEntry{tag} eq $CONSTANT_Double) {
	    &write_u8($cpEntry{val});
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} elsif ($cpEntry{tag} eq $CONSTANT_Long) {
	    &write_u8($cpEntry{val});
	    $i++; ## Ick.  8-byte entries take two constant pool entries
    	} else {
    	    &fatal("Unknown Constant type $cpEntry{tag}!\n");
    	}
    } continue {
    	$i++;
    }
    
    ###
    ### Misc. Class Info
    ###
    
    &write_u2($class{accessFlags});
    &write_u2($class{thisClass});
    &write_u2($class{superClass});
    
    ###
    ### Direct super-interfaces
    ###
    &write_u2($class{interfaceCt});
    
    $i = 0;
    while ($i < $class{interfaceCt}) {
    	&write_u2($class{interfaces}[$i]);
    } continue {
    	$i++;
    }
    
    ###
    ### Fields
    ###
    &write_u2($class{fieldCt});
    
    $i = 0;
    while ($i < $class{fieldCt}) {
    	my %field = %{$class{fields}[$i]};
    	&write_u2($field{accessFlags});
    	&write_u2($field{nameIndex});
    	&write_u2($field{descriptorIndex});
    	&write_u2($field{attributesCt});
    	&writeAttributes($field{attributes});
    } continue {
    	$i++;
    }
    
    ###
    ### Methods
    ###
    &write_u2($class{methodCt});
    
    $i = 0;
    while ($i < $class{methodCt}) {
    	my %method = %{$class{methods}[$i]};
    
    	&write_u2($method{accessFlags});
    	&write_u2($method{nameIndex});
	&write_u2($method{descriptorIndex});
    	&write_u2($method{attributesCt});
    	&writeAttributes($method{attributes});
    } continue {
    	$i++;
    }
    
    ###
    ### Class attributes
    ###
    &write_u2($class{attributesCt});
    &writeAttributes($class{attributes});

    ###
    ### End of .class file
    ###

    return \%class;
}

###
### Integrity check functions
###

sub checkIndex {
    my ($val, $name, $type) = @_;

    # $class is a global 

    if ($val == 0) {
      &fatal("ERROR: Found constant pool index 0 for $name.  (Expecting a CONSTANT_$CONSTANTNames{$type} entry.)");
    }
     
    if ($val >= $class{constantPoolCt}) {
	&fatal("ERROR: $name index for current constant is $val, must be less than $class{constantPoolCt}\n");
    } 

    if (defined($type)) {
	my $actualTag = $class{constantPool}[$val]{tag};
	if ($actualTag != $type) {
	    &fatal("ERROR: $name expects a CONSTANT_$CONSTANTNames{$type} entry at $val, but found a CONSTANT_$CONSTANTNames{$actualTag} entry\n");
	}
    }
}

###
### Read primitives
###

sub read_u8 {
    my $long = 0;

    (read(CLASSIN, $long, 8) == 8) || die ("premature eof in read_u8()\n");

    # my ($b1, $b2, $b3, $b4, $b5, $b6, $b7, $b8) = unpack("CCCCCCCC", $long);
    # return (($b1 << 56) + ($b2 << 48) + ($b3 << 40) + ($b4 << 32)
    #	    + ($b5 << 24) + ($b6 << 16) + ($b7 << 8) + $b4);

    return join(".", unpack("CCCCCCCC", $long));
}

sub read_u4 {
    my $long = 0;
    (read(CLASSIN, $long, 4) == 4) || die ("premature eof in read_u4()\n");
    my ($top, $highmid, $lowmid, $low) = unpack("CCCC", $long);  ## XXX use 'N'
    return ($top * (256*256*256)) + ($highmid * (256*256)) + ($lowmid * 256) + $low;
}

sub read_u2 {
    my $short = 0;
    (read(CLASSIN, $short, 2) == 2) || die ("premature eof in read_u2()\n");
    my ($high, $low) = unpack("CC", $short); ## XXX use 'n'
    #print("read_u2: $high, $low\n");
    return ($high * 256) + $low;
}

sub read_u1 {
    my $byte = 0;
    (read(CLASSIN, $byte, 1) == 1) || die ("premature eof in read_u1()\n");
    my $val = unpack("C", $byte);
    return $val;
}

sub read_n {
    my $byteCt = shift;
    my $foo = '';
    (read(CLASSIN, $foo, $byteCt) == $byteCt) || die ("premature eof in read_n($byteCt)\n");

    return $foo;
}

sub parse_long {
  my $silly = shift;
  my @u8_bits = split(/\./, $silly);

  my $hi = ($u8_bits[0] << 24) + ($u8_bits[1] << 16) + ($u8_bits[2] << 8) + $u8_bits[3];
  my $low = ($u8_bits[4] << 24) + ($u8_bits[5] << 16) + ($u8_bits[6] << 8) + $u8_bits[7];
   
  return sprintf("0x%lx", $low) if ($hi == 0);

  return "0x" . sprintf("%lx", $hi) . sprintf("%8.8lx", $low) 
}

sub parse_float {
    my $intVal = shift;
    
    return "+INF" if ($intVal == 0x7f800000);
    return "-INF" if ($intVal == 0xff800000);
    if ((($intVal >= 0x7f800001) && ($intVal <= 0x7fffffff))
	|| (($intVal >= 0xff800001) && ($intVal <= 0xffffffff))) {
	return "NaN";
    }

    ## Otherwise, convert to a floating point number
    $sign     = (($intVal >> 31) == 0) ?  1 : -1;
    $exponent = (($intVal >> 23) & 0xFF);
    $mantissa = ($exponent == 0) ? ($intVal & 0x7fffff) << 1 : ($intVal & 0x7fffff) | 0x800000;

    return $sign * $mantissa * 2 ** ($exponent - 150);
}

sub parse_double {
    my $intVal = shift;

    if (0) {
      return "+INF" if ($intVal == 0x7f800000);
      return "-INF" if ($intVal == 0xff800000);
      if ((($intVal >= 0x7f800001) && ($intVal <= 0x7fffffff))
	  || (($intVal >= 0xff800001) && ($intVal <= 0xffffffff))) {
	return "NaN";
      }
      
      ## Otherwise, convert to a floating point number
      $sign     = (($intVal >> 31) == 0) ?  1 : -1;
      $exponent = (($intVal >> 23) & 0xFF);
      $mantissa = ($exponent == 0) ? ($intVal & 0x7fffff) << 1 : ($intVal & 0x7fffff) | 0x800000;
      
      return $sign * $mantissa * 2 ** ($exponent - 150);
    }
    return "<fix me>"
}

sub read_utf8 {
    my $byteCt = shift;
    my $utf = '';
    (read(CLASSIN, $utf, $byteCt) == $byteCt) || die ("premature eof in read_utf8($byteCt)\n");

    my $str = unpack("A$byteCt", $utf);
    return $str;
}

sub parseAttribute {
  my ($r_class, $r_attr) = @_; # ref to attr struct
  
  my $attrName = $class{constantPool}[$r_attr->{nameIndex}]->{val};

  if ($attrName eq "Code") {
    my $attr = $r_attr->{rawattr};
    my $offset = 0;
    
    $r_attr->{maxStack} = &grab_u2($offset, $attr); 
    $offset += 2;
    $r_attr->{maxLocals} = &grab_u2($offset, $attr); 
    $offset += 2;
    $r_attr->{codeLength} = &grab_u4($offset, $attr);
    $offset += 4;
    $r_attr->{code} = substr($attr, 8, $r_attr->{codeLength});
    $offset += $r_attr->{codeLength};
    $r_attr->{exceptionTableLen} = &grab_u2($offset, $attr); 
    $offset += 2;
    $r_attr->{exceptionStuff} = substr($attr, $offset, $r_attr->{exceptionTableLen} * 8);
    $offset += $r_attr->{exceptionTableLen} * 8;
    $r_attr->{attributesCount} = &grab_u2($offset, $attr);
    $offset += 2;
    $r_attr->{codeAttrs} = substr($attr, $offset); # remainder of block
    
    ## Double check that we don't read off the end of the block.
    if ($offset > $r_attr->{len}) {
      print "ERROR: malformed code attribute!\n";
    }
  } else {
    ## Don't yet care about this attr.  Feel free to write a parser for it.  :)
  }
}


sub readAttributes {
    my ($r_class, $ct) = @_;
    
    my @attrs = [];

    return undef if ($ct < 1);

    my $i = 0;
    while ($i < $ct) {
	my %attribute;
        $attribute{nameIndex} = &read_u2();
	$attribute{len} = &read_u4();
	$attribute{rawattr} = &read_n($attribute{len});
	
	&checkIndex($attribute{nameIndex}, "Attribute Name", $CONSTANT_Utf8);

	### Parse the attribute (if we care)
	parseAttribute($r_class, \%attribute);

	$attrs[$i] = \%attribute;
    } continue {
	$i++;
    }

    return \@attrs;
}

###
### Write primitives
###

sub write_u4 {
    my $val = shift;
    print CLASSOUT pack("N", $val);
}

sub write_u2 {
    my $val = shift;
    print CLASSOUT pack("n", $val);
}

sub write_u1 {
    my $byte = shift;
    print CLASSOUT pack("C", $byte);
}

sub write_n {
    my $val = shift; 
    print CLASSOUT $val;	# XXX assumes $val."length" is 'n'
}

sub write_utf8 {
    my $utf8 = shift;
    print CLASSOUT $utf8
}

sub writeAttributes {
    my $r_attrs = shift;

    return if (!defined $r_attrs);

    my $i = 0;
    foreach $r_attr (@{$r_attrs}) {
	my %attribute = %{$r_attr};
	&write_u2($attribute{nameIndex});
	&write_u4($attribute{len});
	&write_n($attribute{rawattr});
    }
}

sub fatal {
    print STDERR @_;
    print STDERR "\n";
    exit 11;
}

1;

# eof

%BytecodeNames = 
   (
    0 => "nop",
    1 => "aconst_null",
    2 => "iconst_m1",
    3 => "iconst_0",
    4 => "iconst_1",
    5 => "iconst_2",
    6 => "iconst_3",
    7 => "iconst_4",
    8 => "iconst_5",
    9 => "lconst_0",
    10 => "lconst_1",
    11 => "fconst_0",
    12 => "fconst_1",
    13 => "fconst_2",
    14 => "dconst_0",
    15 => "dconst_1",
    16 => "bipush",
    17 => "sipush",
    18 => "ldc1",
    19 => "ldc2",
    20 => "ldc2w",
    21 => "iload",
    22 => "lload",
    23 => "fload",
    24 => "dload",
    25 => "aload",
    26 => "iload_0",
    27 => "iload_1",
    28 => "iload_2",
    29 => "iload_3",
    30 => "lload_0",
    31 => "lload_1",
    32 => "lload_2",
    33 => "lload_3",
    34 => "fload_0",
    35 => "fload_1",
    36 => "fload_2",
    37 => "fload_3",
    38 => "dload_0",
    39 => "dload_1",
    40 => "dload_2",
    41 => "dload_3",
    42 => "aload_0",
    43 => "aload_1",
    44 => "aload_2",
    45 => "aload_3",
    46 => "iaload",
    47 => "laload",
    48 => "faload",
    49 => "daload",
    50 => "aaload",
    51 => "baload",
    52 => "caload",
    53 => "saload",
    54 => "istore",
    55 => "lstore",
    56 => "fstore",
    57 => "dstore",
    58 => "astore",
    59 => "istore_0",
    60 => "istore_1",
    61 => "istore_2",
    62 => "istore_3",
    63 => "lstore_0",
    64 => "lstore_1",
    65 => "lstore_2",
    66 => "lstore_3",
    67 => "fstore_0",
    68 => "fstore_1",
    69 => "fstore_2",
    70 => "fstore_3",
    71 => "dstore_0",
    72 => "dstore_1",
    73 => "dstore_2",
    74 => "dstore_3",
    75 => "astore_0",
    76 => "astore_1",
    77 => "astore_2",
    78 => "astore_3",
    79 => "iastore",
    80 => "lastore",
    81 => "fastore",
    82 => "dastore",
    83 => "aastore",
    84 => "bastore",
    85 => "castore",
    86 => "sastore",
    87 => "pop",
    88 => "pop2",
    89 => "dup",
    90 => "dup_x1",
    91 => "dup_x2",
    92 => "dup2",
    93 => "dup2_x1",
    94 => "dup2_x2",
    95 => "swap",
    96 => "iadd",
    97 => "ladd",
    98 => "fadd",
    99 => "dadd",
    100 => "isub",
    101 => "lsub",
    102 => "fsub",
    103 => "dsub",
    104 => "imul",
    105 => "lmul",
    106 => "fmul",
    107 => "dmul",
    108 => "idiv",
    109 => "ldiv",
    110 => "fdiv",
    111 => "ddiv",
    112 => "irem",
    113 => "lrem",
    114 => "frem",
    115 => "drem",
    116 => "ineg",
    117 => "lneg",
    118 => "fneg",
    119 => "dneg",
    120 => "ishl",
    121 => "lshl",
    122 => "ishr",
    123 => "lshr",
    124 => "iushr",
    125 => "lushr",
    126 => "iand",
    127 => "land",
    128 => "ior",
    129 => "lor",
    130 => "ixor",
    131 => "lxor",
    132 => "iinc",
    133 => "i2l",
    134 => "i2f",
    135 => "i2d",
    136 => "l2i",
    137 => "l2f",
    138 => "l2d",
    139 => "f2i",
    140 => "f2l",
    141 => "f2d",
    142 => "d2i",
    143 => "d2l",
    144 => "d2f",
    145 => "int2byte",
    146 => "int2char",
    147 => "int2short",
    148 => "lcmp",
    149 => "fcmpl",
    150 => "fcmpg",
    151 => "dcmpl",
    152 => "dcmpg",
    153 => "ifeq",
    154 => "ifne",
    155 => "iflt",
    156 => "ifge",
    157 => "ifgt",
    158 => "ifle",
    159 => "if_icmpeq",
    160 => "if_icmpne",
    161 => "if_icmplt",
    162 => "if_icmpge",
    163 => "if_icmpgt",
    164 => "if_icmple",
    165 => "if_acmpeq",
    166 => "if_acmpne",
    167 => "goto",
    168 => "jsr",
    169 => "ret",
    170 => "tableswitch",
    171 => "lookupswitch",
    172 => "ireturn",
    173 => "lreturn",
    174 => "freturn",
    175 => "dreturn",
    176 => "areturn",
    177 => "return",
    178 => "getstatic",
    179 => "putstatic",
    180 => "getfield",
    181 => "putfield",
    182 => "invokevirtual",
    183 => "invokespecial",
    184 => "invokestatic",
    185 => "invokeinterface",
    187 => "new",
    188 => "newarray",
    189 => "anewarray",
    190 => "arraylength",
    191 => "athrow",
    192 => "checkcast",
    193 => "instanceof",
    194 => "monitorenter",
    195 => "monitorexit",
    196 => "wide",
    197 => "multianewarray",
    198 => "ifnull",
    199 => "ifnonnull",
    200 => "goto_w",
    201 => "jsr_w",
    202 => "breakpoint"
   );
